from .libraries import Tick, TickMath, SwapMath, FullMath, LiquidityMath
from .libraries import Position, SqrtPriceMath, SafeMath

from .libraries.Account import Account
from .libraries.Shared import *
from dataclasses import dataclass


@dataclass
class Slot0:
    ## the current price
    sqrtPriceX96: int
    ## the current tick
    tick: int
    ## the current protocol fee as a percentage of the swap fee taken on withdrawal
    ## represented as an integer denominator (1#x)%
    feeProtocol: int


@dataclass
class ModifyPositionParams:
    ## the address that owns the position
    owner: int
    ## the lower and upper tick of the position
    tickLower: int
    tickUpper: int
    ## any change in liquidity
    liquidityDelta: int


@dataclass
class SwapCache:
    ## the protocol fee for the input token
    feeProtocol: int
    ## liquidity at the beginning of the swap
    liquidityStart: int


## the top level state of the swap, the results of which are recorded in storage at the end
@dataclass
class SwapState:
    ## the amount remaining to be swapped in#out of the input#output asset
    amountSpecifiedRemaining: int
    ## the amount already swapped out#in of the output#input asset
    amountCalculated: int
    ## current sqrt(price)
    sqrtPriceX96: int
    ## the tick associated with the current price
    tick: int
    ## the global fee growth of the input token
    feeGrowthGlobalX128: int
    ## amount of input token paid as protocol fee
    protocolFee: int
    ## the current liquidity in range
    liquidity: int

    ## list of ticks crossed during the swap
    ticksCrossed: list


@dataclass
class StepComputations:
    ## the price at the beginning of the step
    sqrtPriceStartX96: int
    ## the next tick to swap to from the current tick in the swap direction
    tickNext: int
    ## whether tickNext is initialized or not
    initialized: bool
    ## sqrt(price) for the next tick (1#0)
    sqrtPriceNextX96: int
    ## how much is being swapped in in this step
    amountIn: int
    ## how much is being swapped out
    amountOut: int
    ## how much fee is being paid in
    feeAmount: int


@dataclass
class ProtocolFees:
    token0: int
    token1: int


class UniswapPool(Account):

    # Constructor
    def __init__(self, token0, token1, fee, tickSpacing, ledger):
        checkInputTypes(string=(token0, token1), uint24=(fee), int24=(tickSpacing))
        # Contract storage variables
        super().__init__("UniswapPool", [token0, token1], [0, 0])
        self.token0 = token0
        self.token1 = token1
        self.fee = fee
        self.tickSpacing = tickSpacing
        self.maxLiquidityPerTick = Tick.tickSpacingToMaxLiquidityPerTick(tickSpacing)

        # Initialize remaining storage variables
        self.slot0 = Slot0(0, 0, 0)
        self.feeGrowthGlobal0X128 = 0
        self.feeGrowthGlobal1X128 = 0
        self.protocolFees = ProtocolFees(0, 0)
        self.liquidity = 0
        # dict ( int24 => Tick.Info)
        self.ticks = dict()
        self.positions = dict()

        self.ledger = ledger

    ### @dev Common checks for valid tick inputs.
    def checkTicks(tickLower, tickUpper):
        checkInputTypes(int24=(tickLower, tickUpper))
        assert tickLower < tickUpper, "TLU"
        assert tickLower >= TickMath.MIN_TICK, "TLM"
        assert tickUpper <= TickMath.MAX_TICK, "TUM"

    ### @notice Sets the initial price for the pool
    ### @dev Price is represented as a sqrt(amountToken1/amountToken0) Q64.96 value
    ### @param sqrtPriceX96 the initial sqrt price of the pool as a Q64.96
    def initialize(self, sqrtPriceX96):
        checkInputTypes(uint160=(sqrtPriceX96))
        assert self.slot0.sqrtPriceX96 == 0, "AI"

        tick = TickMath.getTickAtSqrtRatio(sqrtPriceX96)

        self.slot0 = Slot0(
            sqrtPriceX96,
            tick,
            0,
        )

    ## @dev Effect some changes to a position
    ## @param params the position details and the change to the position's liquidity to effect
    ## @return position a storage pointer referencing the position with the given owner and tick range
    ## @return amount0 the amount of token0 owed to the pool, negative if the pool should pay the recipient
    ## @return amount1 the amount of token1 owed to the pool, negative if the pool should pay the recipient
    def _modifyPosition(self, params):
        checkInputTypes(
            accounts=(params.owner),
            int24=(params.tickLower, params.tickUpper),
            int128=(params.liquidityDelta),
        )
        UniswapPool.checkTicks(params.tickLower, params.tickUpper)

        # Initialize values
        amount0 = amount1 = 0

        position = self._updatePosition(
            params.owner,
            params.tickLower,
            params.tickUpper,
            params.liquidityDelta,
            self.slot0.tick,
        )

        if params.liquidityDelta != 0:
            if self.slot0.tick < params.tickLower:
                ## current tick is below the passed range; liquidity can only become in range by crossing from left to
                ## right, when we'll need _more_ token0 (it's becoming more valuable) so user must provide it
                amount0 = SqrtPriceMath.getAmount0DeltaHelper(
                    TickMath.getSqrtRatioAtTick(params.tickLower),
                    TickMath.getSqrtRatioAtTick(params.tickUpper),
                    params.liquidityDelta,
                )
            elif self.slot0.tick < params.tickUpper:
                ## current tick is inside the passed range
                amount0 = SqrtPriceMath.getAmount0DeltaHelper(
                    self.slot0.sqrtPriceX96,
                    TickMath.getSqrtRatioAtTick(params.tickUpper),
                    params.liquidityDelta,
                )
                amount1 = SqrtPriceMath.getAmount1DeltaHelper(
                    TickMath.getSqrtRatioAtTick(params.tickLower),
                    self.slot0.sqrtPriceX96,
                    params.liquidityDelta,
                )
                self.liquidity = LiquidityMath.addDelta(
                    self.liquidity, params.liquidityDelta
                )
            else:
                ## current tick is above the passed range; liquidity can only become in range by crossing from right to
                ## left, when we'll need _more_ token1 (it's becoming more valuable) so user must provide it
                amount1 = SqrtPriceMath.getAmount1DeltaHelper(
                    TickMath.getSqrtRatioAtTick(params.tickLower),
                    TickMath.getSqrtRatioAtTick(params.tickUpper),
                    params.liquidityDelta,
                )

        return (position, amount0, amount1)

    ### @dev Gets and updates a position with the given liquidity delta
    ### @param owner the owner of the position
    ### @param tickLower the lower tick of the position's tick range
    ### @param tickUpper the upper tick of the position's tick range
    ### @param tick the current tick, passed to avoid sloads
    def _updatePosition(self, owner, tickLower, tickUpper, liquidityDelta, tick):
        checkInputTypes(
            accounts=(owner),
            int24=(tickLower, tickUpper, tick),
            int128=(liquidityDelta),
        )
        # This will create a position if it doesn't exist
        position = Position.get(self.positions, owner, tickLower, tickUpper)

        # Initialize values
        flippedLower = flippedUpper = False

        ## if we need to update the ticks, do it
        if liquidityDelta != 0:
            flippedLower = Tick.update(
                self.ticks,
                tickLower,
                tick,
                liquidityDelta,
                self.feeGrowthGlobal0X128,
                self.feeGrowthGlobal1X128,
                False,
                self.maxLiquidityPerTick,
            )
            flippedUpper = Tick.update(
                self.ticks,
                tickUpper,
                tick,
                liquidityDelta,
                self.feeGrowthGlobal0X128,
                self.feeGrowthGlobal1X128,
                True,
                self.maxLiquidityPerTick,
            )

        if flippedLower:
            assert tickLower % self.tickSpacing == 0  ## ensure that the tick is spaced
        if flippedUpper:
            assert tickUpper % self.tickSpacing == 0  ## ensure that the tick is spaced

        (feeGrowthInside0X128, feeGrowthInside1X128) = Tick.getFeeGrowthInside(
            self.ticks,
            tickLower,
            tickUpper,
            tick,
            self.feeGrowthGlobal0X128,
            self.feeGrowthGlobal1X128,
        )

        Position.update(
            position, liquidityDelta, feeGrowthInside0X128, feeGrowthInside1X128
        )

        ## clear any tick data that is no longer needed
        if liquidityDelta < 0:
            if flippedLower:
                Tick.clear(self.ticks, tickLower)
            if flippedUpper:
                Tick.clear(self.ticks, tickUpper)
        return position

    ## @notice Adds liquidity for the given recipient/tickLower/tickUpper position
    ## @dev The final amounts calculated are automatically transferred from the swapper
    ## to the pool and vice verse. The amount of token0/token1 due depends
    ## on tickLower, tickUpper, the amount of liquidity, and the current price.
    ## @param recipient The address for which the liquidity will be created
    ## @param tickLower The lower tick of the position in which to add liquidity
    ## @param tickUpper The upper tick of the position in which to add liquidity
    ## @param amount The amount of liquidity to mint
    ## @return amount0 The amount of token0 that was paid to mint the given amount of liquidity.
    ## @return amount1 The amount of token1 that was paid to mint the given amount of liquidity.
    def mint(self, recipient, tickLower, tickUpper, amount):
        checkInputTypes(
            accounts=(recipient), int24=(tickLower, tickUpper), uint128=(amount)
        )
        assert amount > 0

        (_, amount0Int, amount1Int) = self._modifyPosition(
            ModifyPositionParams(recipient, tickLower, tickUpper, amount)
        )

        amount0 = toUint256(abs(amount0Int))
        amount1 = toUint256(abs(amount1Int))

        # Transfer tokens - including safety checks
        self.ledger.transferToken(recipient, self, self.token0, amount0)
        self.ledger.transferToken(recipient, self, self.token1, amount1)

        return (amount0, amount1)

    ## @notice Collects tokens owed to a position
    ## @dev Does not recompute fees earned, which must be done either via mint or burn of any amount of liquidity.
    ## Collect must be called by the position owner. To withdraw only token0 or only token1, amount0Requested or
    ## amount1Requested may be set to zero. To withdraw all tokens owed, caller may pass any value greater than the
    ## actual tokens owed, e.g. type(uint128).max. Tokens owed may be from accumulated swap fees or burned liquidity.
    ## @param recipient The address which should receive the fees collected
    ## @param tickLower The lower tick of the position for which to collect fees
    ## @param tickUpper The upper tick of the position for which to collect fees
    ## @param amount0Requested How much token0 should be withdrawn from the fees owed
    ## @param amount1Requested How much token1 should be withdrawn from the fees owed
    ## @return amount0 The amount of fees collected in token0
    ## @return amount1 The amount of fees collected in token1
    def collect(
        self, recipient, tickLower, tickUpper, amount0Requested, amount1Requested
    ):
        checkInputTypes(
            accounts=(recipient),
            int24=(tickLower, tickUpper),
            uint128=(amount0Requested, amount1Requested),
        )
        # Add this check to prevent creating a new position if the position doesn't exist or it's empty
        position = Position.assertPositionExists(
            self.positions, recipient, tickLower, tickUpper
        )

        amount0 = (
            position.tokensOwed0
            if (amount0Requested > position.tokensOwed0)
            else amount0Requested
        )
        amount1 = (
            position.tokensOwed1
            if (amount1Requested > position.tokensOwed1)
            else amount1Requested
        )

        if amount0 > 0:
            position.tokensOwed0 -= amount0
            self.ledger.transferToken(self, recipient, self.token0, amount0)
        if amount1 > 0:
            position.tokensOwed1 -= amount1
            self.ledger.transferToken(self, recipient, self.token1, amount1)

        return (recipient, tickLower, tickUpper, amount0, amount1)

    ## @notice Burn liquidity from the sender and account tokens owed for the liquidity to the position
    ## @dev Can be used to trigger a recalculation of fees owed to a position by calling with an amount of 0
    ## @dev Fees must be collected separately via a call to #collect
    ## @param tickLower The lower tick of the position for which to burn liquidity
    ## @param tickUpper The upper tick of the position for which to burn liquidity
    ## @param amount How much liquidity to burn
    ## @return amount0 The amount of token0 sent to the recipient
    ## @return amount1 The amount of token1 sent to the recipient
    def burn(self, recipient, tickLower, tickUpper, amount):
        checkInputTypes(
            accounts=(recipient), int24=(tickLower, tickUpper), uint128=(amount)
        )

        # Add check if the position exists - when poking an uninitialized position it can be that
        # getFeeGrowthInside finds a non-initialized tick before Position.update reverts.
        Position.assertPositionExists(self.positions, recipient, tickLower, tickUpper)

        # Added extra recipient input variable to mimic msg.sender
        (position, amount0Int, amount1Int) = self._modifyPosition(
            ModifyPositionParams(recipient, tickLower, tickUpper, -amount)
        )

        # Mimic conversion to uint256
        amount0 = abs(-amount0Int) & (2**256 - 1)
        amount1 = abs(-amount1Int) & (2**256 - 1)

        if amount0 > 0 or amount1 > 0:
            position.tokensOwed0 += amount0
            position.tokensOwed1 += amount1

        return (recipient, tickLower, tickUpper, amount, amount0, amount1)

    ## @notice Swap token0 for token1, or token1 for token0
    ## @dev The tokens are automatically transferred at the end of the swapping function.
    ## @param recipient The address to receive the output of the swap
    ## @param zeroForOne The direction of the swap, true for token0 to token1, false for token1 to token0
    ## @param amountSpecified The amount of the swap, which implicitly configures the swap as exact input (positive), or exact output (negative)
    ## @param sqrtPriceLimitX96 The Q64.96 sqrt price limit. If zero for one, the price cannot be less than this
    ## value after the swap. If one for zero, the price cannot be greater than this value after the swap
    ## @return amount0 The delta of the balance of token0 of the pool, exact when negative, minimum when positive
    ## @return amount1 The delta of the balance of token1 of the pool, exact when negative, minimum when positive
    def swap(self, recipient, zeroForOne, amountSpecified, sqrtPriceLimitX96):
        checkInputTypes(
            accounts=(recipient),
            bool=(zeroForOne),
            int256=(amountSpecified),
            uint160=(sqrtPriceLimitX96),
        )
        assert amountSpecified != 0, "AS"

        slot0Start = self.slot0

        if zeroForOne:
            assert (
                sqrtPriceLimitX96 < slot0Start.sqrtPriceX96
                and sqrtPriceLimitX96 > TickMath.MIN_SQRT_RATIO
            ), "SPL"
        else:
            assert (
                sqrtPriceLimitX96 > slot0Start.sqrtPriceX96
                and sqrtPriceLimitX96 < TickMath.MAX_SQRT_RATIO
            ), "SPL"

        feeProtocol = (
            (slot0Start.feeProtocol % 16)
            if zeroForOne
            else (slot0Start.feeProtocol >> 4)
        )

        cache = SwapCache(feeProtocol, self.liquidity)

        exactInput = amountSpecified > 0

        state = SwapState(
            amountSpecified,
            0,
            slot0Start.sqrtPriceX96,
            slot0Start.tick,
            self.feeGrowthGlobal0X128 if zeroForOne else self.feeGrowthGlobal1X128,
            0,
            cache.liquidityStart,
            [],
        )

        while (
            state.amountSpecifiedRemaining != 0
            and state.sqrtPriceX96 != sqrtPriceLimitX96
        ):
            step = StepComputations(0, 0, 0, 0, 0, 0, 0)
            step.sqrtPriceStartX96 = state.sqrtPriceX96

            (step.tickNext, step.initialized) = self.nextTick(state.tick, zeroForOne)

            ## get the price for the next tick
            step.sqrtPriceNextX96 = TickMath.getSqrtRatioAtTick(step.tickNext)

            ## compute values to swap to the target tick, price limit, or point where input#output amount is exhausted
            if zeroForOne:
                sqrtRatioTargetX96 = (
                    sqrtPriceLimitX96
                    if step.sqrtPriceNextX96 < sqrtPriceLimitX96
                    else step.sqrtPriceNextX96
                )
            else:
                sqrtRatioTargetX96 = (
                    sqrtPriceLimitX96
                    if step.sqrtPriceNextX96 > sqrtPriceLimitX96
                    else step.sqrtPriceNextX96
                )

            (
                state.sqrtPriceX96,
                step.amountIn,
                step.amountOut,
                step.feeAmount,
            ) = SwapMath.computeSwapStep(
                state.sqrtPriceX96,
                sqrtRatioTargetX96,
                state.liquidity,
                state.amountSpecifiedRemaining,
                self.fee,
            )
            if exactInput:
                state.amountSpecifiedRemaining -= step.amountIn + step.feeAmount
                state.amountCalculated = SafeMath.subInts(
                    state.amountCalculated, step.amountOut
                )
            else:
                state.amountSpecifiedRemaining += step.amountOut
                state.amountCalculated = SafeMath.addInts(
                    state.amountCalculated, step.amountIn + step.feeAmount
                )

            ## if the protocol fee is on, calculate how much is owed, decrement feeAmount, and increment protocolFee
            if cache.feeProtocol > 0:
                delta = abs(step.feeAmount // cache.feeProtocol)
                step.feeAmount -= delta
                state.protocolFee += delta & (2**128 - 1)

            ## update global fee tracker
            if state.liquidity > 0:
                state.feeGrowthGlobalX128 += FullMath.mulDiv(
                    step.feeAmount, FixedPoint128_Q128, state.liquidity
                )
                # Addition can overflow in Solidity - mimic it
                state.feeGrowthGlobalX128 = toUint256(state.feeGrowthGlobalX128)

            ## shift tick if we reached the next price
            if state.sqrtPriceX96 == step.sqrtPriceNextX96:
                ## if the tick is initialized, run the tick transition
                ## @dev: here is where we should handle the case of an uninitialized boundary tick
                if step.initialized:
                    liquidityNet = Tick.cross(
                        self.ticks,
                        step.tickNext,
                        state.feeGrowthGlobalX128
                        if zeroForOne
                        else self.feeGrowthGlobal0X128,
                        self.feeGrowthGlobal1X128
                        if zeroForOne
                        else state.feeGrowthGlobalX128,
                    )
                    ## if we're moving leftward, we interpret liquidityNet as the opposite sign
                    ## safe because liquidityNet cannot be type(int128).min
                    if zeroForOne:
                        liquidityNet = -liquidityNet

                    state.liquidity = LiquidityMath.addDelta(
                        state.liquidity, liquidityNet
                    )

                state.tick = (step.tickNext - 1) if zeroForOne else step.tickNext
            elif state.sqrtPriceX96 != step.sqrtPriceStartX96:
                ## recompute unless we're on a lower tick boundary (i.e. already transitioned ticks), and haven't moved
                state.tick = TickMath.getTickAtSqrtRatio(state.sqrtPriceX96)

        ## End of swap loop
        ## update tick
        if state.tick != slot0Start.tick:
            self.slot0.sqrtPriceX96 = state.sqrtPriceX96
            self.slot0.tick = state.tick
        else:
            ## otherwise just update the price
            self.slot0.sqrtPriceX96 = state.sqrtPriceX96

        ## update liquidity if it changed
        if cache.liquidityStart != state.liquidity:
            self.liquidity = state.liquidity

        ## update fee growth global and, if necessary, protocol fees
        ## overflow is acceptable, protocol has to withdraw before it hits type(uint128).max fees

        if zeroForOne:
            self.feeGrowthGlobal0X128 = state.feeGrowthGlobalX128
            if state.protocolFee > 0:
                self.protocolFees.token0 += state.protocolFee
        else:
            self.feeGrowthGlobal1X128 = state.feeGrowthGlobalX128
            if state.protocolFee > 0:
                self.protocolFees.token1 += state.protocolFee

        (amount0, amount1) = (
            (amountSpecified - state.amountSpecifiedRemaining, state.amountCalculated)
            if (zeroForOne == exactInput)
            else (
                state.amountCalculated,
                amountSpecified - state.amountSpecifiedRemaining,
            )
        )

        ## do the transfers and collect payment
        if zeroForOne:
            if amount1 < 0:
                self.ledger.transferToken(self, recipient, self.token1, abs(amount1))
            balanceBefore = self.balances[self.token0]
            self.ledger.transferToken(recipient, self, self.token0, abs(amount0))
            assert balanceBefore + abs(amount0) == self.balances[self.token0], "IIA"
        else:
            if amount0 < 0:
                self.ledger.transferToken(self, recipient, self.token0, abs(amount0))

            balanceBefore = self.balances[self.token1]
            self.ledger.transferToken(recipient, self, self.token1, abs(amount1))
            assert balanceBefore + abs(amount1) == self.balances[self.token1], "IIA"

        return (
            recipient,
            amount0,
            amount1,
            state.sqrtPriceX96,
            state.liquidity,
            state.tick,
        )

    ### @notice Set the denominator of the protocol's % share of the fees
    ### @param feeProtocol0 new protocol fee for token0 of the pool
    ### @param feeProtocol1 new protocol fee for token1 of the pool
    def setFeeProtocol(self, feeProtocol0, feeProtocol1):
        checkInputTypes(uint8=(feeProtocol0, feeProtocol1))
        assert (feeProtocol0 == 0 or (feeProtocol0 >= 4 and feeProtocol0 <= 10)) and (
            feeProtocol1 == 0 or (feeProtocol1 >= 4 and feeProtocol1 <= 10)
        )

        feeProtocolOld = self.slot0.feeProtocol
        feeProtocolNew = feeProtocol0 + (feeProtocol1 << 4)
        # Health check
        checkUInt8(feeProtocolNew)
        self.slot0.feeProtocol = feeProtocolNew
        return (feeProtocolOld % 16, feeProtocolOld >> 4, feeProtocol0, feeProtocol1)

    ### @notice Collect the protocol fee accrued to the pool
    ### @param recipient The address to which collected protocol fees should be sent
    ### @param amount0Requested The maximum amount of token0 to send, can be 0 to collect fees in only token1
    ### @param amount1Requested The maximum amount of token1 to send, can be 0 to collect fees in only token0
    ### @return amount0 The protocol fee collected in token0
    ### @return amount1 The protocol fee collected in token1
    def collectProtocol(self, recipient, amount0Requested, amount1Requested):
        checkInputTypes(
            accounts=(recipient), uint128=(amount0Requested, amount1Requested)
        )
        amount0 = (
            self.protocolFees.token0
            if amount0Requested > self.protocolFees.token0
            else amount0Requested
        )
        amount1 = (
            self.protocolFees.token1
            if amount1Requested > self.protocolFees.token1
            else amount1Requested
        )

        if amount0 > 0:
            if amount0 == self.protocolFees.token0:
                amount0 -= 1  ##ensure that the slot is not cleared, for gas savings
            self.protocolFees.token0 -= amount0
            self.ledger.transferToken(self, recipient, self.token0, amount0)

        if amount1 > 0:
            if amount1 == self.protocolFees.token1:
                amount1 -= 1  ## ensure that the slot is not cleared, for gas savings
            self.protocolFees.token1 -= amount1
            self.ledger.transferToken(self, recipient, self.token1, amount1)

        return recipient, amount0, amount1

    ### @notice It is assumed that the keys are within [MIN_TICK , MAX_TICK], which should always be the case.
    ### We don't run the risk of overshooting tickNext (out of boundaries) as long as ticks (keys) have been initialized
    ### within the boundaries. However, if there is no initialized tick to the left or right we will return the next boundary
    ### Then we need to return the initialized bool to indicate that we are at the boundary and it is not an initalized tick.
    ### @param self The mapping in which to compute the next initialized tick
    ### @param tick The starting tick
    ### @param lte Whether to search for the next initialized tick to the left (less than or equal to the starting tick)
    ### @return Next tick with liquidity to be used in the swap function.
    def nextTick(self, tick, lte):
        checkInputTypes(int24=(tick), bool=(lte))

        keyList = list(self.ticks.keys())

        # If tick doesn't exist in the mapping we fake it (easier than searching for nearest value). This is probably not the
        # best way, but it is a simple and intuitive way to reproduce the behaviour of the logic.
        if not self.ticks.__contains__(tick):
            keyList += [tick]
        sortedKeyList = sorted(keyList)
        indexCurrentTick = sortedKeyList.index(tick)

        if lte:
            # If the current tick is initialized (not faked), we return the current tick
            if self.ticks.__contains__(tick):
                return tick, True
            elif indexCurrentTick == 0:
                # No tick to the left
                return TickMath.MIN_TICK, False
            else:
                nextTick = sortedKeyList[indexCurrentTick - 1]
        else:

            if indexCurrentTick == len(sortedKeyList) - 1:
                # No tick to the right
                return TickMath.MAX_TICK, False
            nextTick = sortedKeyList[indexCurrentTick + 1]

        # Return tick within the boundaries
        return nextTick, True



from .utilities import *
from .poolFixtures import *
from .test_uniswapPool import accounts, ledger
from .UniswapV3PoolSwaps import swapsSnapshot

from ..src.UniswapPool import *
from ..src.libraries.Factory import *


@pytest.fixture(params=[*range(0, 15, 1)])
def TEST_POOLS(request, accounts, ledger):
    poolFixture = request.getfixturevalue("pool{}".format(request.param))
    feeAmount = poolFixture.feeAmount
    tickSpacing = poolFixture.tickSpacing
    pool = UniswapPool(TEST_TOKENS[0], TEST_TOKENS[1], feeAmount, tickSpacing, ledger)
    pool.initialize(poolFixture.startingPrice)
    for position in poolFixture.positions:
        pool.mint(
            accounts[0], position.tickLower, position.tickUpper, position.liquidity
        )
    poolBalance0 = pool.balances[TEST_TOKENS[0]]
    poolBalance1 = pool.balances[TEST_TOKENS[1]]
    return (
        TEST_TOKENS[0],
        TEST_TOKENS[1],
        pool,
        poolBalance0,
        poolBalance1,
        accounts[1],
        poolFixture,
    )


@pytest.fixture
def afterEach(accounts, TEST_POOLS):
    yield
    print("check can burn positions")
    (_, _, pool, _, _, _, poolFixture) = TEST_POOLS
    for position in poolFixture.positions:
        pool.burn(
            accounts[0], position.tickLower, position.tickUpper, position.liquidity
        )
        pool.collect(
            accounts[0],
            position.tickLower,
            position.tickUpper,
            MAX_UINT128,
            MAX_UINT128,
        )


# UniswapV3Pool swap tests


@pytest.mark.usefixtures("afterEach")
def test_uniswap_swaps(TEST_POOLS):
    (_, _, pool, poolBalance0, poolBalance1, recipient, poolFixture) = TEST_POOLS
    print(poolFixture.description)

    if poolFixture.swapTests == None:
        swapTests = DEFAULT_POOL_SWAP_TESTS
    else:
        swapTests = poolFixture.swapTests

    for testCase in swapTests:
        slot0 = pool.slot0
        poolInstance = copy.deepcopy(pool)

        # Get snapshot results
        snapshotIndex = swapsSnapshot.index(
            "UniswapV3Pool swap tests "
            + poolFixture.description
            + " "
            + swapCaseToDescription(testCase)
        )
        dict = swapsSnapshot[snapshotIndex + 1]

        try:
            recipient, amount0, amount1, _, _, _ = executeSwap(
                poolInstance, testCase, recipient
            )
        except AssertionError as msg:
            assert str(msg) == "SPL"
            assert float(dict["poolBalance0"]) == pytest.approx(poolBalance0, rel=1e-12)
            assert float(dict["poolBalance1"]) == pytest.approx(poolBalance1, rel=1e-12)
            decimalPoints = Decimal(dict["poolPriceBefore"]).as_tuple().exponent
            assert float(dict["poolPriceBefore"]) == formatPriceWithPrecision(
                slot0.sqrtPriceX96, -decimalPoints
            )
            assert float(dict["tickBefore"]) == slot0.tick
            continue

        poolBalance0After = poolInstance.balances[TEST_TOKENS[0]]
        poolBalance1After = poolInstance.balances[TEST_TOKENS[1]]
        slot0After = poolInstance.slot0
        liquidityAfter = poolInstance.liquidity
        feeGrowthGlobal0X128 = poolInstance.feeGrowthGlobal0X128
        feeGrowthGlobal1X128 = poolInstance.feeGrowthGlobal1X128

        poolBalance0Delta = poolBalance0After - poolBalance0
        poolBalance1Delta = poolBalance1After - poolBalance1

        ## check all the events were emitted corresponding to balance changes
        assert amount0 == poolBalance0Delta
        assert amount1 == poolBalance1Delta

        if poolBalance0Delta != 0:
            executionPrice = -(poolBalance1Delta / poolBalance0Delta)
        else:
            executionPrice = "-Infinity"

        # Allowing some very small difference due to rounding errors
        assert float(dict["amount0Before"]) == pytest.approx(poolBalance0, rel=1e-12)
        assert float(dict["amount0Delta"]) == pytest.approx(
            poolBalance0Delta, rel=1e-12
        )
        assert float(dict["amount1Before"]) == pytest.approx(poolBalance1, rel=1e-12)
        assert float(dict["amount1Delta"]) == pytest.approx(
            poolBalance1Delta, rel=1e-12
        )
        if dict["executionPrice"] in ["Infinity", "-Infinity", "NaN"]:
            assert executionPrice in ["Infinity", "-Infinity", "NaN"]
        else:
            decimalPoints = Decimal(dict["executionPrice"]).as_tuple().exponent
            assert float(dict["executionPrice"]) == round(
                executionPrice, -decimalPoints
            )
        assert float(dict["feeGrowthGlobal0X128Delta"]) == pytest.approx(
            feeGrowthGlobal0X128, rel=1e-12
        )
        assert float(dict["feeGrowthGlobal1X128Delta"]) == pytest.approx(
            feeGrowthGlobal1X128, rel=1e-12
        )
        # Rounding pool storage variables to the same amount of decimal points as the snapshots
        decimalPoints = Decimal(dict["poolPriceAfter"]).as_tuple().exponent
        assert float(dict["poolPriceAfter"]) == formatPriceWithPrecision(
            slot0After.sqrtPriceX96, -decimalPoints
        )
        decimalPoints = Decimal(dict["poolPriceBefore"]).as_tuple().exponent
        assert float(dict["poolPriceBefore"]) == formatPriceWithPrecision(
            slot0.sqrtPriceX96, -decimalPoints
        )
        assert float(dict["tickAfter"]) == slot0After.tick
        assert float(dict["tickBefore"]) == slot0.tick


def executeSwap(pool, testCase, recipient):
    sqrtPriceLimit = (
        None
        if not testCase.__contains__("sqrtPriceLimit")
        else testCase["sqrtPriceLimit"]
    )
    if testCase.__contains__("exactOut"):
        if testCase["exactOut"]:
            if testCase["zeroForOne"]:
                return swap0ForExact1(
                    pool, testCase["amount1"], recipient, sqrtPriceLimit
                )
            else:
                return swap1ForExact0(
                    pool, testCase["amount0"], recipient, sqrtPriceLimit
                )
        else:
            if testCase["zeroForOne"]:
                return swapExact0For1(
                    pool, testCase["amount0"], recipient, sqrtPriceLimit
                )
            else:
                return swapExact1For0(
                    pool, testCase["amount1"], recipient, sqrtPriceLimit
                )
    else:
        if testCase["zeroForOne"]:
            return swapToLowerPrice(pool, recipient, testCase["sqrtPriceLimit"])
        else:
            return swapToHigherPrice(pool, recipient, testCase["sqrtPriceLimit"])


def swapCaseToDescription(testCase):
    priceClause = (
        " to price " + str(formatPrice(testCase["sqrtPriceLimit"]))
        if testCase.__contains__("sqrtPriceLimit")
        else ""
    )

    if testCase.__contains__("exactOut"):
        if testCase["exactOut"]:
            if testCase["zeroForOne"]:
                return (
                    "swap token0 for exactly "
                    + formatTokenAmount(testCase["amount1"])
                    + " token1"
                    + priceClause
                )
            else:
                return (
                    "swap token1 for exactly "
                    + formatTokenAmount(testCase["amount0"])
                    + " token0"
                    + priceClause
                )
        else:
            if testCase["zeroForOne"]:
                return (
                    "swap exactly "
                    + formatTokenAmount(testCase["amount0"])
                    + " token0 for token1"
                    + priceClause
                )
            else:
                return (
                    "swap exactly "
                    + formatTokenAmount(testCase["amount1"])
                    + " token1 for token0"
                    + priceClause
                )
    else:
        if testCase["zeroForOne"]:
            return "swap token0 for token1" + priceClause
        else:
            return "swap token1 for token0" + priceClause



from .Shared import *
import secrets

# This module is created to mimick blockchain accounts and their balances. For simplification purposess will only
# support the tokens that are initialized in the constructor function. Transfering or receiving other tokens will
# fail (token not in balances dict).

# A random hex address will be assigned to every account when created. This is done to mimic the blockchain address
# and to not store the addressess as just pointers to the account instance. Otherwise issues arise when using
# the account reference values or when we make copies of the Pool in testing.
class Account:
    def __init__(self, name, tokens, balances):
        checkInputTypes(string=(name, *tokens), uint256=(*balances,))
        assert len(tokens) == len(balances)
        # Check uniquness of tokens list
        assert len(set(tokens)) == len(tokens)

        self.name = name

        self.address = secrets.token_hex(40)

        self.tokens = tokens
        self.balances = {}

        # Assign initial balances
        for i in range(len(tokens)):
            self.balances[tokens[i]] = balances[i]

    def updateBalance(self, token, amount):
        checkInputTypes(string=(token), int256=(amount))
        self.balances[token] += amount

        # Check potential overflow/underflow that would happen in solidity
        checkUInt256(self.balances[token])


# The ledger class is used to keep track of all accounts and their balances and to process the transfer of tokens
# between them.
class Ledger:
    def __init__(self, initialAccounts):
        self.accounts = dict()
        for accountParams in initialAccounts:
            self.createAccount(accountParams[0], accountParams[1], accountParams[2])

    def createAccount(self, name, tokens, balances):
        checkInputTypes(string=(name, *tokens), uint256=(balances))
        account = Account(name, tokens, balances)
        assert not self.accounts.__contains__(account.address)
        self.accounts[account.address] = account

    # Add transfer and receive tokens functions.
    def transferToken(self, sender, recipient, token, amount):
        checkInputTypes(account=(recipient), string=(token), uint256=(amount))
        # Get the references to the Account recipient and sender
        if type(recipient) == str:
            recipient = self.getAccountWithAddress(recipient)
        if type(sender) == str:
            sender = self.getAccountWithAddress(sender)

        balanceSenderBefore = sender.balances[token]
        balanceReceiverBefore = recipient.balances[token]

        assert sender.balances[token] >= amount, "Insufficient balance"

        sender.updateBalance(token, -amount)

        self.receiveToken(recipient, token, amount)

        # Transfer health check
        assert sender.balances[token] == balanceSenderBefore - amount
        assert recipient.balances[token] == balanceReceiverBefore + amount

    def receiveToken(self, recipient, token, amount):
        if type(recipient) == str:
            recipient = self.getAccountWithAddress(recipient)

        checkInputTypes(string=(token), uint256=(amount))
        recipient.updateBalance(token, amount)

    def getAccountWithAddress(self, address):
        return self.accounts[address]

    def balanceOf(self, address, token):
        checkInputTypes(string=(address, token))
        return self.accounts[address].balances[token]

    # Force the balance of an account to ease the testing
    def setBalance(self, address, token, amount):
        checkInputTypes(string=(address, token), uint256=amount)
        self.accounts[address].balances[token] = amount



from .utilities import *
from ..src.libraries import TickMath, SqrtPriceMath


def test_fromInput_fails_price_zero():
    print("fails if price is zero")
    tryExceptHandler(
        SqrtPriceMath.getNextSqrtPriceFromInput,
        "",
        0,
        0,
        expandTo18Decimals(1) // 10,
        False,
    )


def test_fromInput_fails_liquidity_zero():
    print("fails if liquidity is zero")
    tryExceptHandler(
        SqrtPriceMath.getNextSqrtPriceFromInput,
        "",
        1,
        0,
        expandTo18Decimals(1) // 10,
        True,
    )


def test_fromInput_fails_input_overflow():
    print("fails if input amount overflows the price")
    price = 2**160 - 1
    liquidity = 1024
    amountIn = 1024
    tryExceptHandler(
        SqrtPriceMath.getNextSqrtPriceFromInput,
        "OF or UF of UINT160",
        price,
        liquidity,
        amountIn,
        False,
    )


def test_fromInput_any_input_overflow():
    print("any input amount cannot underflow the price")
    price = 1
    liquidity = 1
    amountIn = 2**255
    tryExceptHandler(
        SqrtPriceMath.getNextSqrtPriceFromInput,
        "OF or UF of UINT256",
        price,
        liquidity,
        amountIn,
        False,
    )


def test_fromInput_zeroAmount_zeroForOne():
    print("returns input price if amount in is zero and zeroForOne = true")
    price = 2**96
    assert price == SqrtPriceMath.getNextSqrtPriceFromInput(
        price, expandTo18Decimals(1) // 10, 0, True
    )

    print("returns input price if amount in is zero and zeroForOne = false")
    price = 2**96
    assert price == SqrtPriceMath.getNextSqrtPriceFromInput(
        price, expandTo18Decimals(1) // 10, 0, False
    )


def test_fromInput_zeroAmount_notZeroForOne():
    print("returns the minimum price for max inputs")
    sqrtP = 2**160 - 1
    liquidity = TickMath.MAX_UINT128
    maxAmountNoOverflow = TickMath.MAX_UINT256 - ((liquidity << 96) // sqrtP)
    assert 1 == SqrtPriceMath.getNextSqrtPriceFromInput(
        sqrtP, liquidity, maxAmountNoOverflow, True
    )


def test_fromInput_inputAmount_token1():
    print("input amount of 0.1 token1")
    sqrtQ = SqrtPriceMath.getNextSqrtPriceFromInput(
        encodePriceSqrt(1, 1),
        expandTo18Decimals(1),
        int(expandTo18Decimals(1) / 10),
        False,
    )
    assert int(sqrtQ) == 87150978765690771352898345369


def test_fromInput_inputAmount_token0():
    print("input amount of 0.1 token0")
    sqrtQ = SqrtPriceMath.getNextSqrtPriceFromInput(
        encodePriceSqrt(1, 1),
        expandTo18Decimals(1),
        int(expandTo18Decimals(1) / 10),
        True,
    )
    assert int(sqrtQ) == 72025602285694852357767227579


def test_fromInput_amountInMaxUint96_zeroForOne():
    print("amountIn > type(uint96).max and zeroForOne = true")

    sqrtQ = SqrtPriceMath.getNextSqrtPriceFromInput(
        encodePriceSqrt(1, 1), expandTo18Decimals(10), 2**100, True
    )
    assert sqrtQ == 624999999995069620


def test_fromInput_amountInMaxUint96_notZeroForOne():
    print("can return 1 with enough amountIn and zeroForOne = true")
    sqrtQ = SqrtPriceMath.getNextSqrtPriceFromInput(
        encodePriceSqrt(1, 1), 1, int(TickMath.MAX_UINT256 / 2), True
    )
    assert int(sqrtQ) == 1


# Test getNextSqrtPriceFromOutput
def test_fromOutput_fails_zeroPrice():
    print("fails if price is zero")
    tryExceptHandler(
        SqrtPriceMath.getNextSqrtPriceFromOutput,
        "",
        0,
        0,
        expandTo18Decimals(1) // 10,
        False,
    )


def test_fromOutput_fails_liquidityZero():
    print("fails if price is zero")
    tryExceptHandler(
        SqrtPriceMath.getNextSqrtPriceFromOutput,
        "",
        1,
        0,
        expandTo18Decimals(1) // 10,
        True,
    )


def test_fromOutput_fails_equalOutputReserves_token0():
    print("fails if output amount is exactly the virtual reserves of token0")
    price = 20282409603651670423947251286016
    liquidity = 1024
    amountOut = 4
    tryExceptHandler(
        SqrtPriceMath.getNextSqrtPriceFromOutput, "", price, liquidity, amountOut, False
    )


def test_fromOutput_fails_greaterOutputReserves_token0():
    print("fails if output amount is greater than virtual reserves of token0")
    price = 20282409603651670423947251286016
    liquidity = 1024
    amountOut = 5
    tryExceptHandler(
        SqrtPriceMath.getNextSqrtPriceFromOutput, "", price, liquidity, amountOut, False
    )


def test_fromOutput_fails_greaterOutputReserves_token1():
    print("fails if output amount is greater than virtual reserves of token1")
    price = 20282409603651670423947251286016
    liquidity = 1024
    amountOut = 262145
    tryExceptHandler(
        SqrtPriceMath.getNextSqrtPriceFromOutput, "", price, liquidity, amountOut, True
    )


def test_fromOutput_fails_equalOutputReserves_token1():
    print("fails if output amount is exactly the virtual reserves of token1")
    price = 20282409603651670423947251286016
    liquidity = 1024
    amountOut = 262144
    tryExceptHandler(
        SqrtPriceMath.getNextSqrtPriceFromOutput, "", price, liquidity, amountOut, True
    )


def test_fromOutput_output_lessThanReserves_token1():
    print("succeeds if output amount is just less than the virtual reserves of token1")
    price = 20282409603651670423947251286016
    liquidity = 1024
    amountOut = 262143
    sqrtQ = SqrtPriceMath.getNextSqrtPriceFromOutput(price, liquidity, amountOut, True)
    assert int(sqrtQ) == 77371252455336267181195264


def test_fromOutput_puzzlingEchidnaTest():
    print("puzzling echidna test")
    price = 20282409603651670423947251286016
    liquidity = 1024
    amountOut = 4
    tryExceptHandler(
        SqrtPriceMath.getNextSqrtPriceFromOutput, "", price, liquidity, amountOut, False
    )


def test_fromOutput_zeroAmountIn_zeroForOne():
    print("returns input price if amount in is zero and zeroForOne = true")
    price = encodePriceSqrt(1, 1)
    sqrtQ = SqrtPriceMath.getNextSqrtPriceFromOutput(
        price, int(expandTo18Decimals(1) / 10), 0, True
    )
    assert int(sqrtQ) == price


def test_fromOutput_zeroAmountIn_notZeroForOne():
    print("returns input price if amount in is zero and zeroForOne = false")
    price = encodePriceSqrt(1, 1)
    sqrtQ = SqrtPriceMath.getNextSqrtPriceFromOutput(
        price, int(expandTo18Decimals(1) / 10), 0, False
    )
    assert int(sqrtQ) == price


def test_fromOutput_outputAmount_token1_notZeroForOne():
    print("output amount of 0.1 token1")
    sqrtQ = SqrtPriceMath.getNextSqrtPriceFromOutput(
        encodePriceSqrt(1, 1),
        expandTo18Decimals(1),
        int(expandTo18Decimals(1) / 10),
        False,
    )
    assert int(sqrtQ) == 88031291682515930659493278152


def test_fromOutput_outputAmount_token1_zeroForOne():
    print("output amount of 0.1 token1")
    sqrtQ = SqrtPriceMath.getNextSqrtPriceFromOutput(
        encodePriceSqrt(1, 1),
        expandTo18Decimals(1),
        int(expandTo18Decimals(1) / 10),
        True,
    )
    assert int(sqrtQ) == 71305346262837903834189555302


def test_fails_impossibleAmountOut_zeroForOne():
    print("reverts if amountOut is impossible in zero for one direction")
    # Reverts in divRoundingUp
    tryExceptHandler(
        SqrtPriceMath.getNextSqrtPriceFromOutput,
        "OF or UF of UINT256",
        encodePriceSqrt(1, 1),
        1,
        int(TickMath.MAX_UINT256 / 2),
        True,
    )


def test_fails_impossibleAmountOut_notZeroForOne():
    print("reverts if amountOut is impossible in zero for one direction")
    tryExceptHandler(
        SqrtPriceMath.getNextSqrtPriceFromOutput,
        "",
        encodePriceSqrt(1, 1),
        1,
        int(TickMath.MAX_UINT256 / 2),
        False,
    )


# getAmount0Delta


def test_getAmount0Delta_zeroLiquidity():
    print("returns if liquidity is zero")
    sqrtQ = SqrtPriceMath.getAmount0Delta(
        encodePriceSqrt(1, 1), encodePriceSqrt(2, 1), 0, True
    )
    assert sqrtQ == 0


def test_getAmount0Delta_equalPrices():
    print("returns 0 if prices are equal")
    sqrtQ = SqrtPriceMath.getAmount0Delta(
        encodePriceSqrt(1, 1), encodePriceSqrt(1, 1), 0, True
    )
    assert sqrtQ == 0


def test_getAmount0Delta_returnsAmount1():
    print("returns 0.1 amount1 for price of 1 to 1.21")

    amount0 = SqrtPriceMath.getAmount0Delta(
        encodePriceSqrt(1, 1), encodePriceSqrt(121, 100), expandTo18Decimals(1), True
    )
    assert int(amount0) == 90909090909090910


def test_getAmount0Delta_priceOverflow():
    print("works for prices that overflow")
    amount0Up = SqrtPriceMath.getAmount0Delta(
        encodePriceSqrt(2**90, 1),
        encodePriceSqrt(2**96, 1),
        expandTo18Decimals(1),
        True,
    )
    amount0Down = SqrtPriceMath.getAmount0Delta(
        encodePriceSqrt(2**90, 1),
        encodePriceSqrt(2**96, 1),
        expandTo18Decimals(1),
        False,
    )

    assert int(amount0Up) == int(amount0Down) + 1


# getAmount1Delta


def test_getAmount1Delta_zeroLiquidity():
    print("returns if liquidity is zero")
    sqrtQ = SqrtPriceMath.getAmount1Delta(
        encodePriceSqrt(1, 1), encodePriceSqrt(2, 1), 0, True
    )
    assert sqrtQ == 0


def test_getAmount1Delta_equalPrices():
    print("returns 0 if prices are equal")
    sqrtQ = SqrtPriceMath.getAmount1Delta(
        encodePriceSqrt(1, 1), encodePriceSqrt(1, 1), 0, True
    )
    assert sqrtQ == 0


def test_getAmount1Delta_returnsAmount1():
    print("returns 0.1 amount1 for price of 1 to 1.21")
    print(encodePriceSqrt(1, 1))
    print(encodePriceSqrt(121, 100))

    amount0 = SqrtPriceMath.getAmount1Delta(
        encodePriceSqrt(1, 1), encodePriceSqrt(121, 100), expandTo18Decimals(1), True
    )
    assert int(amount0) == 100000000000000000

    amount1RoundedDown = SqrtPriceMath.getAmount1Delta(
        encodePriceSqrt(1, 1), encodePriceSqrt(121, 100), expandTo18Decimals(1), False
    )

    assert int(amount1RoundedDown) == int(amount0) - 1


# Swap Computation
def test_swap():
    print("swap computation")
    ## getNextSqrtPriceInvariants(1025574284609383690408304870162715216695788925244,50015962439936049619261659728067971248,406,true)
    sqrtP = 1025574284609383690408304870162715216695788925244
    liquidity = 50015962439936049619261659728067971248
    zeroForOne = True
    amountIn = 406

    sqrtQ = SqrtPriceMath.getNextSqrtPriceFromInput(
        sqrtP, liquidity, amountIn, zeroForOne
    )
    assert int(sqrtQ) == 1025574284609383582644711336373707553698163132913

    amount0Delta = SqrtPriceMath.getAmount0Delta(int(sqrtQ), sqrtP, liquidity, True)
    assert int(amount0Delta) == 406



from .utilities import *
from ..src.libraries import SwapMath, SqrtPriceMath

# ComputeSwapStep


def test_exactAmountIn_capped_notOneForZero():
    print("exact amount in that gets capped at price target in one for zero")
    price = encodePriceSqrt(1, 1)
    priceTarget = encodePriceSqrt(101, 100)
    liquidity = expandTo18Decimals(2)
    amount = expandTo18Decimals(1)
    fee = 600
    zeroForOne = False
    (sqrtQ, amountIn, amountOut, feeAmount) = SwapMath.computeSwapStep(
        price, priceTarget, liquidity, amount, fee
    )

    assert int(amountIn) == 9975124224178055
    assert int(feeAmount) == 5988667735148
    assert int(amountOut) == 9925619580021728
    assert (amountIn + feeAmount) < amount, "Entire amount used"

    priceAfterWholeInputAmount = SqrtPriceMath.getNextSqrtPriceFromInput(
        price, liquidity, amount, zeroForOne
    )

    assert int(sqrtQ) == priceTarget, "price not capped at price target"
    assert int(sqrtQ) < int(
        priceAfterWholeInputAmount
    ), "price not less than price after whole input amount"


def test_exactAmountOut_capped_notOneForZero():
    print("exact amount out that gets capped at price target in one for zero")
    price = encodePriceSqrt(1, 1)
    priceTarget = encodePriceSqrt(101, 100)
    liquidity = expandTo18Decimals(2)
    amount = expandTo18Decimals(1) * -1
    fee = 600
    zeroForOne = False

    (sqrtQ, amountIn, amountOut, feeAmount) = SwapMath.computeSwapStep(
        price, priceTarget, liquidity, amount, fee
    )

    assert int(amountIn) == 9975124224178055
    assert int(feeAmount) == 5988667735148
    assert int(amountOut) == 9925619580021728
    assert int(amountOut) < amount * -1, "entire amount out is not returned"

    priceAfterWholeOutputAmount = SqrtPriceMath.getNextSqrtPriceFromOutput(
        price, liquidity, amount * -1, zeroForOne
    )

    assert int(sqrtQ) == priceTarget, "price is capped at price target"
    assert (
        int(sqrtQ) < priceAfterWholeOutputAmount
    ), "price is less than price after whole output amount"


def test_exactAmount_fullySpent_notzeroForOne():
    print("exact amount in that is fully spent in one for zero")
    price = encodePriceSqrt(1, 1)
    priceTarget = encodePriceSqrt(1000, 100)
    liquidity = expandTo18Decimals(2)
    amount = expandTo18Decimals(1)
    fee = 600
    zeroForOne = False

    (sqrtQ, amountIn, amountOut, feeAmount) = SwapMath.computeSwapStep(
        price, priceTarget, liquidity, amount, fee
    )

    assert int(amountIn) == 999400000000000000
    assert int(feeAmount) == 600000000000000
    assert int(amountOut) == 666399946655997866
    assert (int(amountIn) + int(feeAmount)) == amount, "entire amount is used"

    print(type(feeAmount))
    priceAfterWholeInputAmountLessFee = SqrtPriceMath.getNextSqrtPriceFromInput(
        price, liquidity, amount - feeAmount, zeroForOne
    )

    assert int(sqrtQ) < priceTarget, "price does not reach price target"
    assert int(sqrtQ) == int(
        priceAfterWholeInputAmountLessFee
    ), "price is equal to price after whole input amount"


def test_exactAmountOut_fullyReceived_notZeroForOne():
    print("exact amount out that is fully received in one for zero")
    price = encodePriceSqrt(1, 1)
    priceTarget = encodePriceSqrt(1000, 100)
    liquidity = expandTo18Decimals(2)
    amount = expandTo18Decimals(1) * -1
    fee = 600
    zeroForOne = False

    (sqrtQ, amountIn, amountOut, feeAmount) = SwapMath.computeSwapStep(
        price, priceTarget, liquidity, amount, fee
    )

    assert int(amountIn) == 2000000000000000000
    assert int(feeAmount) == 1200720432259356
    assert int(amountOut) == amount * -1

    priceAfterWholeOutputAmount = SqrtPriceMath.getNextSqrtPriceFromOutput(
        price, liquidity, amount * -1, zeroForOne
    )

    assert int(sqrtQ) < priceTarget, "price does not reach price target"
    assert int(sqrtQ) == int(
        priceAfterWholeOutputAmount
    ), "price is equal to price after whole output amount"


def test_exactAmountOut_capped_amountOut():
    print("amount out is capped at the desired amount out")
    (sqrtQ, amountIn, amountOut, feeAmount) = SwapMath.computeSwapStep(
        417332158212080721273783715441582,
        1452870262520218020823638996,
        159344665391607089467575320103,
        -1,
        1,
    )

    assert int(amountIn) == 1
    assert int(feeAmount) == 1
    assert int(amountOut) == 1  ## would be 2 if not capped
    assert int(sqrtQ) == 417332158212080721273783715441581


# This test doesn't make much sence since amountOut is 1 and not 0 since because removed rounding
# However, checking that the other calculations are correct
def test_targetOne_partialInputAmount():
    print("target price of 1 uses partial input amount")
    (sqrtQ, amountIn, amountOut, feeAmount) = SwapMath.computeSwapStep(
        2, 1, 1, 3915081100057732413702495386755767, 1
    )

    assert int(amountIn) == 39614081257132168796771975168
    assert int(feeAmount) == 39614120871253040049813
    assert int(amountIn) + int(feeAmount) <= 3915081100057732413702495386755767
    assert int(amountOut) == 0
    assert int(sqrtQ) == 1


def test_allInput_asFee():
    print("entire input amount taken as fee")
    (sqrtQ, amountIn, amountOut, feeAmount) = SwapMath.computeSwapStep(
        2413, 79887613182836312, 1985041575832132834610021537970, 10, 1872
    )
    assert amountIn == 0
    assert feeAmount == 10
    assert amountOut == 0
    assert sqrtQ == 2413


# This test doesn't make much sence since amountOut is 1 and not 0 since because removed rounding
# However, checking that the other calculations are correct
def test_insuffLiquid_zero_exactOutput():
    print(
        "handles intermediate insufficient liquidity in zero for one exact output case"
    )
    sqrtP = 20282409603651670423947251286016
    sqrtPTarget = sqrtP * 11 // 10
    liquidity = 1024
    ## virtual reserves of one are only 4
    ## https://www.wolframalpha.com/input/?i=1024+%2F+%2820282409603651670423947251286016+%2F+2**96%29
    amountRemaining = -4
    feePips = 3000
    (sqrtQ, amountIn, amountOut, feeAmount) = SwapMath.computeSwapStep(
        sqrtP, sqrtPTarget, liquidity, amountRemaining, feePips
    )

    assert amountOut == 0
    assert sqrtQ == sqrtPTarget
    assert amountIn == 26215
    assert feeAmount == 79



from .utilities import *
from .test_uniswapPool import ledger

from ..src.UniswapPool import *
from ..src.libraries import TickMath

# Instead of testing tickBitmap library, we test the UniswapPool ticks python dict and
# nextTick functionality, which should be equivalent.


# nextTick

# Pass tickSpacing as a workaround since originally TickBitmap is a library but now
# it is part of the UniswapPool.
def initializePoolWithMockTicks(tickSpacing, ledger):
    # To mimic tickBitmaps we need to create a dict of initialized ticks (ticks)
    pool = UniswapPool("A", "B", 1, tickSpacing, ledger)
    insertUninitializedTickstoMapping(
        pool.ticks, [-200, -55, -4, 70, 78, 84, 139, 240, 535]
    )
    return pool


# lte == False


def test_lte_eq_false_notLte_notLte(ledger):
    print("returns tick to right if at initialized tick")
    pool = initializePoolWithMockTicks(1, ledger)
    (next, initialized) = pool.nextTick(78, False)

    assert next == 84
    assert initialized == True


def test_tickRight_ifAtInitializedTick_notLte(ledger):
    print("returns tick to right if at initialized tick")
    pool = initializePoolWithMockTicks(1, ledger)
    (next, initialized) = pool.nextTick(-55, False)

    assert next == -4
    assert initialized == True


def test_tickRight_nonInit_notLte(ledger):
    print("returns the tick directly to the right")
    pool = initializePoolWithMockTicks(1, ledger)
    (next, initialized) = pool.nextTick(77, False)

    assert next == 78
    assert initialized == True


def test_tickRight_init_notLte(ledger):
    print("returns the tick directly to the right")
    pool = initializePoolWithMockTicks(1, ledger)
    (next, initialized) = pool.nextTick(-56, False)

    assert next == -55
    assert initialized == True


# Now we are not limited to 256 visibility from the current tick
def test_nextInit_rightBoundary_notLte(ledger):
    print("returns the next words initialized tick if on the right boundary")
    pool = initializePoolWithMockTicks(1, ledger)
    (next, initialized) = pool.nextTick(255, False)

    assert next == 535
    assert initialized == True


def test_nextInit_leftBoundary_notLte(ledger):
    print("returns the next words initialized tick if on the left boundary")
    pool = initializePoolWithMockTicks(1, ledger)
    (next, initialized) = pool.nextTick(-257, False)

    assert next == -200
    assert initialized == True


def test_nextInit_notLte(ledger):
    print("returns the next initialized tick from the next word")
    pool = initializePoolWithMockTicks(1, ledger)
    insertUninitializedTickstoMapping(pool.ticks, [340])
    (next, initialized) = pool.nextTick(328, False)

    assert next == 340
    assert initialized == True


# Now we are not limited to 256 visibility from the current tick
def test_return_rightNoWordBoundaries_notLte(ledger):
    print("returns the next tick regardless of 256 word boundaries")
    pool = initializePoolWithMockTicks(1, ledger)
    (next, initialized) = pool.nextTick(508, False)
    assert next == 535
    assert initialized == True

    (next, initialized) = pool.nextTick(255, False)
    assert next == 535
    assert initialized == True

    (next, initialized) = pool.nextTick(383, False)
    assert next == 535
    assert initialized == True


# lte == True


def test_sameTick_ifInitialized_lte(ledger):
    print("returns the same tick if initialized tick")
    pool = initializePoolWithMockTicks(1, ledger)
    (next, initialized) = pool.nextTick(78, True)

    assert next == 78
    assert initialized == True


def test_tickLeft_ifNotInitialized_lte(ledger):
    print("returns tick directly to the left of input tick if not initialized")
    pool = initializePoolWithMockTicks(1, ledger)
    (next, initialized) = pool.nextTick(79, True)

    assert next == 78
    assert initialized == True


# Now we are not limited to 256 visibility from the current tick
def test_crossWordBoundary_lte(ledger):
    print("can exceed a word boundary")
    pool = initializePoolWithMockTicks(1, ledger)
    (next, initialized) = pool.nextTick(258, True)
    assert next == 240
    assert initialized == True

    (next, initialized) = pool.nextTick(256, True)
    assert next == 240
    assert initialized == True

    (next, initialized) = pool.nextTick(72, True)
    assert next == 70
    assert initialized == True


# We are now returning the TICK_MAX
def test_return_leftBoundary_lte(ledger):
    print("returns the next tick if at the right boundary")
    pool = initializePoolWithMockTicks(1, ledger)
    (next, initialized) = pool.nextTick(-257, True)

    assert next == TickMath.MIN_TICK
    assert initialized == False


# Now we are not limited to 256 visibility from the current tick - returning highest tick
def test_nextInit_leftBoundary_lte(ledger):
    print("returns the next words initialized tick if on the right boundary")
    pool = initializePoolWithMockTicks(1, ledger)

    (next, initialized) = pool.nextTick(1023, True)
    assert next == 535
    assert initialized == True

    (next, initialized) = pool.nextTick(900, True)
    assert next == 535
    assert initialized == True


def test_wordBoundaryInitialized(ledger):
    print("boundary is initialized")
    pool = initializePoolWithMockTicks(1, ledger)
    insertUninitializedTickstoMapping(pool.ticks, [329])

    (next, initialized) = pool.nextTick(456, True)
    assert next == 329
    assert initialized == True



from .utilities import *


@dataclass
class PoolTestCase:
    description: str
    feeAmount: int
    tickSpacing: int
    startingPrice: int
    positions: list
    swapTests: list


@dataclass
class Position:
    tickLower: int
    tickUpper: int
    liquidity: int


@pytest.fixture
def pool0():
    return PoolTestCase(
        description="low fee, 1:1 price, 2e18 max range liquidity",
        feeAmount=FeeAmount.LOW,
        tickSpacing=TICK_SPACINGS[FeeAmount.LOW],
        startingPrice=encodePriceSqrt(1, 1),
        positions=[
            Position(
                tickLower=getMinTick(TICK_SPACINGS[FeeAmount.LOW]),
                tickUpper=getMaxTick(TICK_SPACINGS[FeeAmount.LOW]),
                liquidity=expandTo18Decimals(2),
            ),
        ],
        swapTests=None,
    )


@pytest.fixture
def pool1():
    return PoolTestCase(
        description="medium fee, 1:1 price, 2e18 max range liquidity",
        feeAmount=FeeAmount.MEDIUM,
        tickSpacing=TICK_SPACINGS[FeeAmount.MEDIUM],
        startingPrice=encodePriceSqrt(1, 1),
        positions=[
            Position(
                tickLower=getMinTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                tickUpper=getMaxTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                liquidity=expandTo18Decimals(2),
            ),
        ],
        swapTests=None,
    )


@pytest.fixture
def pool2():
    return PoolTestCase(
        description="high fee, 1:1 price, 2e18 max range liquidity",
        feeAmount=FeeAmount.HIGH,
        tickSpacing=TICK_SPACINGS[FeeAmount.HIGH],
        startingPrice=encodePriceSqrt(1, 1),
        positions=[
            Position(
                tickLower=getMinTick(TICK_SPACINGS[FeeAmount.HIGH]),
                tickUpper=getMaxTick(TICK_SPACINGS[FeeAmount.HIGH]),
                liquidity=expandTo18Decimals(2),
            ),
        ],
        swapTests=None,
    )


@pytest.fixture
def pool3():
    return PoolTestCase(
        description="medium fee, 10:1 price, 2e18 max range liquidity",
        feeAmount=FeeAmount.MEDIUM,
        tickSpacing=TICK_SPACINGS[FeeAmount.MEDIUM],
        startingPrice=encodePriceSqrt(10, 1),
        positions=[
            Position(
                tickLower=getMinTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                tickUpper=getMaxTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                liquidity=expandTo18Decimals(2),
            ),
        ],
        swapTests=None,
    )


@pytest.fixture
def pool4():
    return PoolTestCase(
        description="medium fee, 1:10 price, 2e18 max range liquidity",
        feeAmount=FeeAmount.MEDIUM,
        tickSpacing=TICK_SPACINGS[FeeAmount.MEDIUM],
        startingPrice=encodePriceSqrt(1, 10),
        positions=[
            Position(
                tickLower=getMinTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                tickUpper=getMaxTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                liquidity=expandTo18Decimals(2),
            ),
        ],
        swapTests=None,
    )


@pytest.fixture
def pool5():
    return PoolTestCase(
        description="medium fee, 1:1 price, 0 liquidity, all liquidity around current price",
        feeAmount=FeeAmount.MEDIUM,
        tickSpacing=TICK_SPACINGS[FeeAmount.MEDIUM],
        startingPrice=encodePriceSqrt(1, 1),
        positions=[
            Position(
                tickLower=getMinTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                tickUpper=-TICK_SPACINGS[FeeAmount.MEDIUM],
                liquidity=expandTo18Decimals(2),
            ),
            Position(
                tickLower=TICK_SPACINGS[FeeAmount.MEDIUM],
                tickUpper=getMaxTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                liquidity=expandTo18Decimals(2),
            ),
        ],
        swapTests=None,
    )


@pytest.fixture
def pool6():
    return PoolTestCase(
        description="medium fee, 1:1 price, additional liquidity around current price",
        feeAmount=FeeAmount.MEDIUM,
        tickSpacing=TICK_SPACINGS[FeeAmount.MEDIUM],
        startingPrice=encodePriceSqrt(1, 1),
        positions=[
            Position(
                tickLower=getMinTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                tickUpper=getMaxTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                liquidity=expandTo18Decimals(2),
            ),
            Position(
                tickLower=getMinTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                tickUpper=-TICK_SPACINGS[FeeAmount.MEDIUM],
                liquidity=expandTo18Decimals(2),
            ),
            Position(
                tickLower=TICK_SPACINGS[FeeAmount.MEDIUM],
                tickUpper=getMaxTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                liquidity=expandTo18Decimals(2),
            ),
        ],
        swapTests=None,
    )


@pytest.fixture
def pool7():
    return PoolTestCase(
        description="low fee, large liquidity around current price (stable swap)",
        feeAmount=FeeAmount.LOW,
        tickSpacing=TICK_SPACINGS[FeeAmount.LOW],
        startingPrice=encodePriceSqrt(1, 1),
        positions=[
            Position(
                tickLower=-TICK_SPACINGS[FeeAmount.LOW],
                tickUpper=TICK_SPACINGS[FeeAmount.LOW],
                liquidity=expandTo18Decimals(2),
            )
        ],
        swapTests=None,
    )


@pytest.fixture
def pool8():
    return PoolTestCase(
        description="medium fee, token0 liquidity only",
        feeAmount=FeeAmount.MEDIUM,
        tickSpacing=TICK_SPACINGS[FeeAmount.MEDIUM],
        startingPrice=encodePriceSqrt(1, 1),
        positions=[
            Position(
                tickLower=0,
                tickUpper=2000 * TICK_SPACINGS[FeeAmount.MEDIUM],
                liquidity=expandTo18Decimals(2),
            )
        ],
        swapTests=None,
    )


@pytest.fixture
def pool9():
    return PoolTestCase(
        description="medium fee, token1 liquidity only",
        feeAmount=FeeAmount.MEDIUM,
        tickSpacing=TICK_SPACINGS[FeeAmount.MEDIUM],
        startingPrice=encodePriceSqrt(1, 1),
        positions=[
            Position(
                tickLower=-2000 * TICK_SPACINGS[FeeAmount.MEDIUM],
                tickUpper=0,
                liquidity=expandTo18Decimals(2),
            )
        ],
        swapTests=None,
    )


@pytest.fixture
def pool10():
    return PoolTestCase(
        description="close to max price",
        feeAmount=FeeAmount.MEDIUM,
        tickSpacing=TICK_SPACINGS[FeeAmount.MEDIUM],
        startingPrice=encodePriceSqrt(2**127, 1),
        positions=[
            Position(
                tickLower=getMinTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                tickUpper=getMaxTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                liquidity=expandTo18Decimals(2),
            )
        ],
        swapTests=None,
    )


# This pool with such low starting price (or someething else) ends up with a bigger
# than normal rounding error when comparing amount0before (and possibly others)
@pytest.fixture
def pool11():
    return PoolTestCase(
        description="close to min price",
        feeAmount=FeeAmount.MEDIUM,
        tickSpacing=TICK_SPACINGS[FeeAmount.MEDIUM],
        startingPrice=encodePriceSqrt(1, 2**127),
        positions=[
            Position(
                tickLower=getMinTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                tickUpper=getMaxTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                liquidity=expandTo18Decimals(2),
            )
        ],
        swapTests=None,
    )


@pytest.fixture
def pool12():
    return PoolTestCase(
        description="max full range liquidity at 1:1 price with default fee",
        feeAmount=FeeAmount.MEDIUM,
        tickSpacing=TICK_SPACINGS[FeeAmount.MEDIUM],
        startingPrice=encodePriceSqrt(1, 1),
        positions=[
            Position(
                tickLower=getMinTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                tickUpper=getMaxTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                liquidity=getMaxLiquidityPerTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
            )
        ],
        swapTests=None,
    )


@pytest.fixture
def pool13():
    return PoolTestCase(
        description="initialized at the max ratio",
        feeAmount=FeeAmount.MEDIUM,
        tickSpacing=TICK_SPACINGS[FeeAmount.MEDIUM],
        startingPrice=MAX_SQRT_RATIO - 1,
        positions=[
            Position(
                tickLower=getMinTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                tickUpper=getMaxTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                liquidity=expandTo18Decimals(2),
            )
        ],
        swapTests=None,
    )


@pytest.fixture
def pool14():
    return PoolTestCase(
        description="initialized at the min ratio",
        feeAmount=FeeAmount.MEDIUM,
        tickSpacing=TICK_SPACINGS[FeeAmount.MEDIUM],
        startingPrice=MIN_SQRT_RATIO,
        positions=[
            Position(
                tickLower=getMinTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                tickUpper=getMaxTick(TICK_SPACINGS[FeeAmount.MEDIUM]),
                liquidity=expandTo18Decimals(2),
            )
        ],
        swapTests=None,
    )


@dataclass
class SWAP_TEST:
    zeroForOne: bool
    exactOut: bool
    amount0: int


DEFAULT_POOL_SWAP_TESTS = [
    {
        ## swap large amounts in/out
        "zeroForOne": True,
        "exactOut": False,
        "amount0": expandTo18Decimals(1),
    },
    {
        "zeroForOne": False,
        "exactOut": False,
        "amount1": expandTo18Decimals(1),
    },
    {
        "zeroForOne": True,
        "exactOut": True,
        "amount1": expandTo18Decimals(1),
    },
    {
        "zeroForOne": False,
        "exactOut": True,
        "amount0": expandTo18Decimals(1),
    },
    ## swap large amounts in/out with a price limit
    {
        "zeroForOne": True,
        "exactOut": False,
        "amount0": expandTo18Decimals(1),
        "sqrtPriceLimit": encodePriceSqrt(50, 100),
    },
    {
        "zeroForOne": False,
        "exactOut": False,
        "amount1": expandTo18Decimals(1),
        "sqrtPriceLimit": encodePriceSqrt(200, 100),
    },
    {
        "zeroForOne": True,
        "exactOut": True,
        "amount1": expandTo18Decimals(1),
        "sqrtPriceLimit": encodePriceSqrt(50, 100),
    },
    {
        "zeroForOne": False,
        "exactOut": True,
        "amount0": expandTo18Decimals(1),
        "sqrtPriceLimit": encodePriceSqrt(200, 100),
    },
    # ## swap small amounts in/out
    # Also, it is only off in extreme pools, not in the ones closer to normal functioning.
    # {
    #   "zeroForOne": True,
    #   "exactOut": False,
    #   "amount0": 1000,
    # },
    # {
    #   "zeroForOne": False,
    #   "exactOut": False,
    #   "amount1": 1000,
    # },
    {
        "zeroForOne": True,
        "exactOut": True,
        "amount1": 1000,
    },
    {
        "zeroForOne": False,
        "exactOut": True,
        "amount0": 1000,
    },
    # ## swap arbitrary input to price
    {
        "sqrtPriceLimit": encodePriceSqrt(5, 2),
        "zeroForOne": False,
    },
    {
        "sqrtPriceLimit": encodePriceSqrt(2, 5),
        "zeroForOne": True,
    },
    {
        "sqrtPriceLimit": encodePriceSqrt(5, 2),
        "zeroForOne": True,
    },
    {
        "sqrtPriceLimit": encodePriceSqrt(2, 5),
        "zeroForOne": False,
    },
]



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="uniswapV3Python",
    version="1.0.0",
    author="Chainflip Labs",
    author_email="albert@chainflip.io",
    description="Pythonized Uniswap V3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/chainflip-io/python-uniswap-v3",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=[
        "uniswapV3Python",
        "uniswapV3Python.src",
        "uniswapV3Python.src.libraries",
        "uniswapV3Python.tests",
    ],
    python_requires=">=3.7",
)



import sys, traceback, math, copy
from decimal import *
import pytest

from ..src.libraries.Shared import *

TEST_TOKENS = ["Token0", "Token1"]


def getMinTick(tickSpacing):
    return math.ceil(MIN_TICK / tickSpacing) * tickSpacing


def getMaxTick(tickSpacing):
    return math.floor(MAX_TICK / tickSpacing) * tickSpacing


def getMaxLiquidityPerTick(tickSpacing):
    denominator = (getMaxTick(tickSpacing) - getMinTick(tickSpacing)) // tickSpacing + 1
    return (2**128 - 1) // denominator


@dataclass
class FeeAmount:
    LOW: int = 500
    MEDIUM: int = 3000
    HIGH: int = 10000


TICK_SPACINGS = {FeeAmount.LOW: 10, FeeAmount.MEDIUM: 60, FeeAmount.HIGH: 200}


def encodePriceSqrt(reserve1, reserve0):
    # Workaround to get the same numbers as JS
    # This ratio doesn't output the same number as in JS using big number. This causes some
    # disparities in the expected results. Full ratios (1,1), (2,1) ...
    # Forcing values obtained by bigNumber.js when ratio is not exact.
    if reserve1 == 121 and reserve0 == 100:
        return 87150978765690771352898345369
    elif reserve1 == 101 and reserve0 == 100:
        return 79623317895830914510487008059
    elif reserve1 == 1 and reserve0 == 10:
        return 25054144837504793118650146401
    elif reserve1 == 1 and reserve0 == 2**127:
        return 6085630636
    elif reserve1 == 2**127 and reserve0 == 1:
        return 1033437718471923701407239276819587054334136928048
    else:
        return int(math.sqrt(reserve1 / reserve0) * 2**96)


def expandTo18Decimals(number):
    # Converting to int because python cannot shl on a float
    return int(number * 10**18)


# @dev This function will handle reverts (aka assert failures) in the tests. However, in python there is no revert
# as in the blockchain. So we will create a hard copy of the current pool and call the same method there.
def tryExceptHandler(fcn, assertMessage, *args):

    reverted = False

    try:
        # reference to object
        pool = fcn.__self__
        fcnName = fcn.__name__

        # hard copy to prevent state changes in the pool
        poolCopy = copy.deepcopy(pool)

        try:
            fcn = getattr(poolCopy, fcnName)
        except AttributeError:
            assert "Function not found in pool: " + fcnName
    except:
        # e.g. case when swapExact1ForZero is expected to revert
        print(
            "Non-pool class function passed - expect pool to be copied as part of the call"
        )

    try:
        fcn(*args)
    except AssertionError as msg:
        reverted = True
        _, _, tb = sys.exc_info()
        traceback.print_tb(tb)  # Fixed format
        tb_info = traceback.extract_tb(tb)
        filename, line, func, text = tb_info[-1]

        print("An error occurred on line {} in statement {}".format(line, text))
        if str(msg) != assertMessage:
            print(
                "Reverted succesfully but not for the expected reason. \n Expected: '"
                + str(assertMessage)
                + "' but got: '"
                + str(msg)
                + "'"
            )
            assert False
        print("Succesful revert")

    if not reverted:
        print("Failed to revert: " + assertMessage)
        assert False


def getPositionKey(address, lowerTick, upperTick):
    return hash((address, lowerTick, upperTick))


def getLimitPositionKey(address, tick, isToken0):
    return hash((address, tick, isToken0))


### POOL SWAPS ###
def swapExact0For1(pool, amount, recipient, sqrtPriceLimit):
    sqrtPriceLimitX96 = (
        sqrtPriceLimit
        if sqrtPriceLimit != None
        else getSqrtPriceLimitX96(TEST_TOKENS[0])
    )
    return swap(pool, TEST_TOKENS[0], [amount, 0], recipient, sqrtPriceLimitX96)


def swap0ForExact1(pool, amount, recipient, sqrtPriceLimit):
    sqrtPriceLimitX96 = (
        sqrtPriceLimit
        if sqrtPriceLimit != None
        else getSqrtPriceLimitX96(TEST_TOKENS[0])
    )
    return swap(pool, TEST_TOKENS[0], [0, amount], recipient, sqrtPriceLimitX96)


def swapExact1For0(pool, amount, recipient, sqrtPriceLimit):
    sqrtPriceLimitX96 = (
        sqrtPriceLimit
        if sqrtPriceLimit != None
        else getSqrtPriceLimitX96(TEST_TOKENS[1])
    )
    return swap(pool, TEST_TOKENS[1], [amount, 0], recipient, sqrtPriceLimitX96)


def swap1ForExact0(pool, amount, recipient, sqrtPriceLimit):
    sqrtPriceLimitX96 = (
        sqrtPriceLimit
        if sqrtPriceLimit != None
        else getSqrtPriceLimitX96(TEST_TOKENS[1])
    )
    return swap(pool, TEST_TOKENS[1], [0, amount], recipient, sqrtPriceLimitX96)


def swapToLowerPrice(pool, recipient, sqrtPriceLimit):
    return pool.swap(recipient, True, MAX_INT256, sqrtPriceLimit)


def swapToHigherPrice(pool, recipient, sqrtPriceLimit):
    return pool.swap(recipient, False, MAX_INT256, sqrtPriceLimit)


def swap(pool, inputToken, amounts, recipient, sqrtPriceLimitX96):
    [amountIn, amountOut] = amounts
    exactInput = amountOut == 0
    amount = amountIn if exactInput else amountOut

    if inputToken == TEST_TOKENS[0]:
        if exactInput:
            checkInt128(amount)
            return pool.swap(recipient, True, amount, sqrtPriceLimitX96)
        else:
            checkInt128(-amount)
            return pool.swap(recipient, True, -amount, sqrtPriceLimitX96)
    else:
        if exactInput:
            checkInt128(amount)
            return pool.swap(recipient, False, amount, sqrtPriceLimitX96)
        else:
            checkInt128(-amount)
            return pool.swap(recipient, False, -amount, sqrtPriceLimitX96)


def getSqrtPriceLimitX96(inputToken):
    if inputToken == TEST_TOKENS[0]:
        return MIN_SQRT_RATIO + 1
    else:
        return MAX_SQRT_RATIO - 1


################


def formatPrice(price):
    fraction = (price / (2**96)) ** 2
    return formatAsInSnapshot(fraction)


def formatTokenAmount(amount):
    fraction = amount / (10**18)
    return formatAsInSnapshot(fraction)


def formatAsInSnapshot(number):
    # To match snapshot formatting
    precision = int(f"{number:e}".split("e")[-1])
    # For token we want 4 extra decimals of precision
    if precision >= 0:
        precision = 4
    else:
        precision = -precision + 4

    return format(number, "." + str(precision) + "f")


def formatPriceWithPrecision(price, precision):
    fraction = (price / (2**96)) ** 2
    return round(fraction, precision)



from .test_uniswapPool import ledger

from ..src.libraries.Factory import *
from .utilities import *

TEST_ADDRESSES = [
    "0x1000000000000000000000000000000000000000",
    "0x2000000000000000000000000000000000000000",
]


def test_initial_fees():
    print("initial enabled fee amounts")
    factory = Factory()
    assert factory.feeAmountTickSpacing[FeeAmount.LOW] == TICK_SPACINGS[FeeAmount.LOW]
    assert (
        factory.feeAmountTickSpacing[FeeAmount.MEDIUM]
        == TICK_SPACINGS[FeeAmount.MEDIUM]
    )
    assert factory.feeAmountTickSpacing[FeeAmount.HIGH] == TICK_SPACINGS[FeeAmount.HIGH]


def createAndCheck_pool(factory, tokens, feeAmount, tickSpacing, ledger):
    print("create and check pool")

    pool = factory.createPool(tokens[0], tokens[1], feeAmount, ledger)

    tryExceptHandler(
        factory.createPool,
        "Pool already exists",
        tokens[0],
        tokens[1],
        feeAmount,
        ledger,
    )
    tryExceptHandler(
        factory.createPool,
        "Pool already exists",
        tokens[1],
        tokens[0],
        feeAmount,
        ledger,
    )

    assert pool.token0 == TEST_ADDRESSES[0], "pool token0"
    assert pool.token1 == TEST_ADDRESSES[1], "pool token1"
    assert pool.fee == feeAmount, "pool fee"
    assert pool.tickSpacing == tickSpacing, "pool tick spacing"


# createPool
def test_lowFeePool(ledger):
    print("succeds for low fee pool")
    createAndCheck_pool(
        Factory(), TEST_ADDRESSES, FeeAmount.LOW, TICK_SPACINGS[FeeAmount.LOW], ledger
    )


def test_mediumFeePool(ledger):
    print("succeds for medium fee pool")
    createAndCheck_pool(
        Factory(),
        TEST_ADDRESSES,
        FeeAmount.MEDIUM,
        TICK_SPACINGS[FeeAmount.MEDIUM],
        ledger,
    )


def test_highFeePool(ledger):
    print("succeds for high fee pool")
    createAndCheck_pool(
        Factory(), TEST_ADDRESSES, FeeAmount.HIGH, TICK_SPACINGS[FeeAmount.HIGH], ledger
    )


def test_tokensReverse(ledger):
    print("succeds for tokens in reverse order")
    createAndCheck_pool(
        Factory(),
        TEST_ADDRESSES[::-1],
        FeeAmount.MEDIUM,
        TICK_SPACINGS[FeeAmount.MEDIUM],
        ledger,
    )


def test_fails_tokenAequalB(ledger):
    print("fails if token a is 0 or token b is 0")
    tryExceptHandler(
        Factory().createPool, "", TEST_ADDRESSES[0], "0", FeeAmount.LOW, ledger
    )
    tryExceptHandler(
        Factory().createPool, "", "0", TEST_ADDRESSES[0], FeeAmount.LOW, ledger
    )
    tryExceptHandler(Factory().createPool, "", "0", "0", FeeAmount.LOW, ledger)


def test_fails_feeAmountNotEnabled(ledger):
    print("fails if fee amount is not enabled")
    tryExceptHandler(
        Factory().createPool,
        "Fee amount not supported",
        TEST_ADDRESSES[0],
        TEST_ADDRESSES[1],
        250,
        ledger,
    )


# setO
def test_fails_feeTooGreat():
    print("fails if fee is too great")
    tryExceptHandler(Factory().enableFeeAmount, "", 1000000, 10)


def test_fails_tickSpacingTooSmall():
    print("fails if tick spacing is too small")
    tryExceptHandler(Factory().enableFeeAmount, "", 500, 0)


def test_fails_tickSpacingTooLarge():
    print("fails if tick spacing is too large")
    tryExceptHandler(Factory().enableFeeAmount, "", 500, 16834)


def test_fails_alreadyInitialized():
    print("fails if already initialized")
    factory = Factory()
    factory.enableFeeAmount(100, 5)
    tryExceptHandler(Factory().enableFeeAmount, "", 100, 0)


def test_setFee():
    print("sets the fee amount in the mapping")
    factory = Factory()
    factory.enableFeeAmount(100, 5)
    assert factory.feeAmountTickSpacing[100] == 5


def test_enablePoolCreation(ledger):
    factory = Factory()
    factory.enableFeeAmount(250, 15)
    createAndCheck_pool(factory, TEST_ADDRESSES, 250, 15, ledger)



from .Shared import *
from . import TickMath

### @title Overflow and underflow math operations.
### @notice Contains methods for doing math operations that revert on overflow or underflow for minimal gas cost.Mimic solidity overflow and underflow check as in some cases the check is a safeguard.

### @notice Returns x + y, reverts if sum overflows uint256
### @param x The augend
### @param y The addend
### @return z The sum of x and y
def add(x, y):
    checkInputTypes(uint256=(x, y))
    z = x + y
    assert z <= TickMath.MAX_UINT256
    return z


### @notice Returns x - y, reverts if underflows
### @param x The minuend
### @param y The subtrahend
### @return z The difference of x and y
def sub(x, y):
    checkInputTypes(uint256=(x, y))
    z = x - y
    assert z >= 0
    return z


### @notice Returns x * y, reverts if overflows
### @param x The multiplicand
### @param y The multiplier
### @return z The product of x and y
def mul(x, y):
    checkInputTypes(uint256=(x, y))
    z = x * y
    assert z <= TickMath.MAX_UINT256
    return z


### @notice Returns x + y, reverts if overflows or underflows
### @param x The augend
### @param y The addend
### @return z The sum of x and y
def addInts(x, y):
    checkInputTypes(int256=(x, y))
    z = x + y
    assert z >= TickMath.MIN_INT256 and z <= TickMath.MAX_UINT256
    return z


### @notice Returns x - y, reverts if overflows or underflows
### @param x The minuend
### @param y The subtrahend
### @return z The difference of x and y
def subInts(x, y):
    checkInputTypes(int256=(x, y))
    z = x - y
    assert z >= TickMath.MIN_INT256 and z <= TickMath.MAX_UINT256
    return z



from . import TickMath, LiquidityMath, SafeMath
import math
from .Shared import *


### @notice Derives max liquidity per tick from given tick spacing
### @dev Executed within the pool constructor
### @param tickSpacing The amount of required tick separation, realized in multiples of `tickSpacing`
###     e.g., a tickSpacing of 3 requires ticks to be initialized every 3rd tick i.e., ..., -6, -3, 0, 3, 6, ...
### @return The max liquidity per tick
def tickSpacingToMaxLiquidityPerTick(tickSpacing):
    checkInt24(tickSpacing)
    minTick = math.ceil(TickMath.MIN_TICK / tickSpacing) * tickSpacing
    maxTick = math.floor(TickMath.MAX_TICK / tickSpacing) * tickSpacing
    assert abs(maxTick) >= abs(minTick)  # Health check
    numTicks = ((maxTick - minTick) // tickSpacing) + 1
    assert numTicks <= 2**24 - 1  # Health check
    return TickMath.MAX_UINT128 // numTicks


### @notice Retrieves fee growth data
### @param self The mapping containing all tick information for initialized ticks
### @param tickLower The lower tick boundary of the position
### @param tickUpper The upper tick boundary of the position
### @param tickCurrent The current tick
### @param feeGrowthGlobal0X128 The all-time global fee growth, per unit of liquidity, in token0
### @param feeGrowthGlobal1X128 The all-time global fee growth, per unit of liquidity, in token1
### @return feeGrowthInside0X128 The all-time fee growth in token0, per unit of liquidity, inside the position's tick boundaries
### @return feeGrowthInside1X128 The all-time fee growth in token1, per unit of liquidity, inside the position's tick boundaries
def getFeeGrowthInside(
    self, tickLower, tickUpper, tickCurrent, feeGrowthGlobal0X128, feeGrowthGlobal1X128
):
    checkInputTypes(
        dict=self,
        int24=(tickLower, tickUpper, tickCurrent),
        uint256=(feeGrowthGlobal0X128, feeGrowthGlobal1X128),
    )

    # Assumption that the key (tick) exists
    lower = self[tickLower]
    upper = self[tickUpper]

    ## calculate fee growth below
    if tickCurrent >= tickLower:
        feeGrowthBelow0X128 = lower.feeGrowthOutside0X128
        feeGrowthBelow1X128 = lower.feeGrowthOutside1X128
    else:
        feeGrowthBelow0X128 = feeGrowthGlobal0X128 - lower.feeGrowthOutside0X128
        feeGrowthBelow1X128 = feeGrowthGlobal1X128 - lower.feeGrowthOutside1X128

    ## calculate fee growth above
    if tickCurrent < tickUpper:
        feeGrowthAbove0X128 = upper.feeGrowthOutside0X128
        feeGrowthAbove1X128 = upper.feeGrowthOutside1X128
    else:
        feeGrowthAbove0X128 = feeGrowthGlobal0X128 - upper.feeGrowthOutside0X128
        feeGrowthAbove1X128 = feeGrowthGlobal1X128 - upper.feeGrowthOutside1X128

    # Mimic overflow from solidity
    feeGrowthInside0X128 = toUint256(
        feeGrowthGlobal0X128 - feeGrowthBelow0X128 - feeGrowthAbove0X128
    )
    feeGrowthInside1X128 = toUint256(
        feeGrowthGlobal1X128 - feeGrowthBelow1X128 - feeGrowthAbove1X128
    )
    return (feeGrowthInside0X128, feeGrowthInside1X128)


### @notice Updates a tick and returns true if the tick was flipped from initialized to uninitialized, or vice versa
### @param self The mapping containing all tick information for initialized ticks
### @param tick The tick that will be updated
### @param tickCurrent The current tick
### @param liquidityDelta A new amount of liquidity to be added (subtracted) when tick is crossed from left to right (right to left)
### @param feeGrowthGlobal0X128 The all-time global fee growth, per unit of liquidity, in token0
### @param feeGrowthGlobal1X128 The all-time global fee growth, per unit of liquidity, in token1
### @param secondsPerLiquidityCumulativeX128 The all-time seconds per max(1, liquidity) of the pool
### @param time The current block timestamp cast to a uint32
### @param upper true for updating a position's upper tick, or false for updating a position's lower tick
### @param maxLiquidity The maximum liquidity allocation for a single tick
### @return flipped Whether the tick was flipped from initialized to uninitialized, or vice versa
def update(
    self,
    tick,
    tickCurrent,
    liquidityDelta,
    feeGrowthGlobal0X128,
    feeGrowthGlobal1X128,
    upper,
    maxLiquidity,
):
    checkInputTypes(
        dict=self,
        int24=(tick, tickCurrent),
        int128=liquidityDelta,
        uint256=(feeGrowthGlobal0X128, feeGrowthGlobal1X128),
        bool=upper,
        uint128=maxLiquidity,
    )
    # Tick might not exist - create it. Make sure tick is not created unless it is then initialized with liquidityDelta > 0
    if not self.__contains__(tick):
        assert liquidityDelta > 0, "Avoid creating empty tick"
        insertUninitializedTickstoMapping(self, [tick])

    info = self[tick]

    liquidityGrossBefore = info.liquidityGross
    liquidityGrossAfter = LiquidityMath.addDelta(liquidityGrossBefore, liquidityDelta)

    assert liquidityGrossAfter <= maxLiquidity, "LO"

    flipped = (liquidityGrossAfter == 0) != (liquidityGrossBefore == 0)

    if liquidityGrossBefore == 0:
        ## by convention, we assume that all growth before a tick was initialized happened _below_ the tick
        if tick <= tickCurrent:
            info.feeGrowthOutside0X128 = feeGrowthGlobal0X128
            info.feeGrowthOutside1X128 = feeGrowthGlobal1X128

    info.liquidityGross = liquidityGrossAfter

    ## when the lower (upper) tick is crossed left to right (right to left), liquidity must be added (removed)
    if upper:
        info.liquidityNet = SafeMath.subInts(info.liquidityNet, liquidityDelta)
        checkInt128(info.liquidityNet)
    else:
        info.liquidityNet = SafeMath.addInts(info.liquidityNet, liquidityDelta)
        checkInt128(info.liquidityNet)

    # No longer require flip to signal if it has been initialized but it is needed for when it is cleared
    return flipped


### @notice Clears tick data
### @param self The mapping containing all initialized tick information for initialized ticks
### @param tick The tick that will be cleared
def clear(self, tick):
    checkInputTypes(dict=self, int24=tick)
    # Assumption that the key (tick) exists (it should)
    del self[tick]


### @notice Transitions to next tick as needed by price movement
### @param self The mapping containing all tick information for initialized ticks
### @param tick The destination tick of the transition
### @param feeGrowthGlobal0X128 The all-time global fee growth, per unit of liquidity, in token0
### @param feeGrowthGlobal1X128 The all-time global fee growth, per unit of liquidity, in token1
### @param secondsPerLiquidityCumulativeX128 The current seconds per liquidity
### @param time The current block.timestamp
### @return liquidityNet The amount of liquidity added (subtracted) when tick is crossed from left to right (right to left)
def cross(
    tickBitmap,
    tick,
    feeGrowthGlobal0X128,
    feeGrowthGlobal1X128,
):
    checkInputTypes(
        dict=tickBitmap,
        int24=tick,
        uint256=(feeGrowthGlobal0X128, feeGrowthGlobal1X128),
    )
    ## TickBitMap is passed by reference so all changes will be applied to the original dict
    ## TickBitMap = dict(uint256 tick => TickInfo)
    info = tickBitmap[tick]
    info.feeGrowthOutside0X128 = feeGrowthGlobal0X128 - info.feeGrowthOutside0X128
    info.feeGrowthOutside1X128 = feeGrowthGlobal1X128 - info.feeGrowthOutside1X128
    liquidityNet = info.liquidityNet
    return liquidityNet



from .Shared import checkInputTypes
from ..UniswapPool import *

## @title Pool factory
## @notice Deploys Uniswap V3 pools and manages ownership and control over pool protocol fees
class Factory:
    def __init__(self):
        self.feeAmountTickSpacing = {500: 10, 3000: 60, 10000: 200}
        self.getPool = []

    ## @notice Creates a pool for the given two tokens and fee
    ## @param tokenA One of the two tokens in the desired pool
    ## @param tokenB The other of the two tokens in the desired pool
    ## @param fee The desired fee for the pool
    ## @param ledger Reference to the ledger
    ## @dev tokenA and tokenB may be passed in either order: token0/token1 or token1/token0. tickSpacing is retrieved
    ## from the fee. The call will revert if the pool already exists, the fee is invalid, or the token arguments
    ## are invalid.
    ## @return pool The address of the newly created pool
    def createPool(self, tokenA, tokenB, fee, ledger):
        checkInputTypes(string=(tokenA, tokenB), uint24=(fee))
        assert tokenA != tokenB

        (token0, token1) = (tokenA, tokenB) if tokenA < tokenB else (tokenB, tokenA)
        assert token0 != "0"
        assert self.feeAmountTickSpacing.__contains__(fee), "Fee amount not supported"
        assert self.feeAmountTickSpacing[fee] != 0
        tickSpacing = self.feeAmountTickSpacing[fee]

        if [token0, token1, fee] not in self.getPool:
            self.getPool.append([token0, token1, fee])
        else:
            assert False, "Pool already exists"

        pool = UniswapPool(token0, token1, fee, tickSpacing, ledger)

        return pool

    ## @notice Enables a fee amount with the given tickSpacing
    ## @dev Fee amounts may never be removed once enabled
    ## @param fee The fee amount to enable, denominated in hundredths of a bip (i.e. 1e-6)
    ## @param tickSpacing The spacing between ticks to be enforced for all pools created with the given fee amount
    def enableFeeAmount(self, fee, tickSpacing):
        checkInputTypes(uint24=(fee), int24=(tickSpacing))
        assert fee < 1000000
        ## tick spacing is capped at 16384 to prevent the situation where tickSpacing is so large that
        ## TickBitmap#nextInitializedTickWithinOneWord overflows int24 container from a valid tick
        ## 16384 ticks represents a >5x price change with ticks of 1 bips
        assert tickSpacing > 0 and tickSpacing < 16384
        assert (
            not self.feeAmountTickSpacing.__contains__(fee)
            or self.feeAmountTickSpacing[fee] == 0
        )
        self.feeAmountTickSpacing[fee] = tickSpacing
